/*************************************************************
 *
 *  main.c - Arbitrary-dimensional diffusion-limited aggregation
 *
 *  Copyright (C) 2000-2005  Mark J. Stock, mstock@umich.edu
 * 
 *  This file is part of dla-nd.
 *
 *  dla-nd is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  dla-nd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dla-nd; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Revision information
 *
 * 2003-12-03 MJS  v0.0  copied part3d source, begin work
 * 2003-12-04 MJS  v0.1  improved order over O(N^2)
 * 2003-12-04 MJS  v0.2  new output, correct aggregation
 * 2004-01-20 MJS  v0.3  added directionality, reorganized
 * 2004-02-27 MJS  v0.4  corrected directionality, added bias,
 *                       added stickiness
 * 2004-08-26 MJS  v0.5  added grip, reorganized parameters,
 *                       added obj output
 * 2004-08-31 MJS  v0.6  upgraded to arbitrary dimensions
 * 2005-02-24 MJS  v0.7  native png output, adaptive trees, no gifs
 * 2005-02-26 MJS  v0.8  accurate bulk motion, experimental rotex
 * 2005-09-02 MJS  v0.9  timing, stats and temp output, much faster
 *
 *********************************************************** */


#include "structs.h"

int run_sim(fileprop_ptr,sim_ptr,cell_ptr);
int diffuse_new_particle(sim_ptr,cell_ptr,FLOAT*,FLOAT*);
void pick_point_on_sphere(FLOAT*,FLOAT*,FLOAT,int,FLOAT*);
FLOAT find_dist_to_closest(cell_ptr,FLOAT*,FLOAT,particle_ptr,particle_ptr*);
FLOAT find_dist_to_farthest(cell_ptr,FLOAT*,FLOAT);
void sum_tipward_mass(cell_ptr);
void zero_all_masses(cell_ptr);
void increment_flags(cell_ptr);
void zero_flags(cell_ptr);
FLOAT vec_dot(FLOAT*,FLOAT*);
FLOAT vec_length(FLOAT*);
FLOAT vec_length_sq(FLOAT*);
void gaussian_rand(FLOAT*,FLOAT*);
FLOAT solve_quartic(FLOAT*);


int main(int argc,char **argv) {

   FILEP file_props;		// initialize struct for file props
   fileprop_ptr file = &file_props;
   SIMP sim_props;		// initialize struct for simulation props
   sim_ptr sim = &sim_props;
   CELL top_cell;		// initialize struct for ndtree
   cell_ptr top = &top_cell;

   // set some important pointers
   file->out = &file->out_field;
   sim->ff2 = &sim->density_field2;
   sim->ff3 = &sim->density_field3;

   /* Parse command-line args */
   parse_args(argc,argv,file);

   /* Initialize the top-level cell and participating particles */
   initialize_system(file,sim,top);

   /* Do the time integration */
   run_sim(file,sim,top);

   fprintf(stderr,"\nDone.\n");
   exit(0);
}



/*
 *  Run the time integration of the flowfield simulation
 */
int run_sim(fileprop_ptr file,sim_ptr sim,cell_ptr top) {

   int d,retval;
   FLOAT this_step_time,total_time,start[DIM],vel[DIM];

   sim->next_output_step = sim->output_step;

   // open the stats file if necessary
   if (file->write_stats) {
      file->statsfp = fopen(file->stats_file,"w");
      if (file->statsfp==NULL) {
         fprintf(stderr,"Could not open output file %s\n",file->stats_file);
         fflush(stderr);
         exit(0);
      }
   }

   // make a good tree
   top = build_new_tree(file,sim,top);

   // start the timer
   sim->last_tics = clock();
   total_time = 0.0;

   // loop over max number of particles
   for (sim->step=1;sim->step<sim->particle_cnt+1;sim->step++) {

      // Write some stuff
      if (sim->step%1000 == 0) {
         fprintf(stdout,"\nStep %d, next print %g, %d particles within %g radius\n",sim->step,sim->next_output_step,top->num,sim->overall_rad);
         fflush(stdout);
      }
      // write particle count and cell utilization
      // if (sim->step%10000 == 0) {
         // fprintf(stdout,"  %d particles\n",top->num);
         // for (i=0; i<sim->max_levels; i++) sim->cell_count[i] = 0;
         // count_cells(sim,top);
         // fprintf(stdout,"  ");
         // for (i=0; i<sim->max_levels; i++) fprintf(stdout,"%d ",sim->cell_count[i]);
         // fprintf(stdout,"cells\n  utilization ");
         // for (i=0; i<sim->max_levels; i++) fprintf(stdout,"%g ",sim->cell_count[i]/pow(8,i));
         // fprintf(stdout,"\n");
      // }
      // if (sim->step%100 == 0) {
         // fprintf(stdout,"."); fflush(stdout);
      // }

      // check current step vs. next_output_step, write output if necessary
      if (sim->step == sim->next_output_step) {
         // stop the timer, save the time
         sim->tics = clock();
         this_step_time = ((FLOAT)sim->tics-(FLOAT)sim->last_tics)/CLOCKS_PER_SEC;
         if (this_step_time < 0.0) this_step_time += 4.2949673e+09/(FLOAT)CLOCKS_PER_SEC;
         total_time += this_step_time;

         // write the output
         fprintf(stdout,"\n  writing output step %d\n",sim->next_output_index);
         write_output(file,sim,top);
         sim->next_output_step += sim->output_step;
         sim->next_output_index++;

         // write a line to the stats file
         if (file->write_stats) {
            fprintf(file->statsfp,"%d %d %g %g %g %g %g\n",sim->step,sim->output_step,
                    sim->overall_rad,
                    (FLOAT)sim->total_particles_tried/(FLOAT)sim->output_step,
                    (FLOAT)sim->total_steps_taken/(FLOAT)sim->output_step,
                    (FLOAT)this_step_time/(FLOAT)sim->output_step,
                    total_time);
            fflush(file->statsfp);
         } else {
            fprintf(stdout,"step time %g, total time %g\n",this_step_time,total_time);
         }

         // reset the counters
         sim->total_particles_tried = 0;
         sim->total_steps_taken = 0;

         // start a new timer
         sim->last_tics = clock();
      }

      // place a new particle
      // diffuse_new_particle(sim,top,0.0,0.0,0.0);
      for (d=0;d<DIM;d++) start[d]=0.0;

      // for Wittmann and Stanley's fake surface tension, zero flags
      if (sim->use_stubborn) zero_flags(top);

      if (sim->use_bulk_vel) {
         // fprintf(stdout,"Using bulk vel, stick=%g, pen=%g\n",sim->stickiness,sim->penetration);
         retval = diffuse_new_particle(sim,top,start,sim->bulk_vel);
      } else {
         // fprintf(stdout,"No bulk vel, stick=%g, pen=%g\n",sim->stickiness,sim->penetration);
         for (d=0;d<DIM;d++) vel[d]=0.0;
         retval = diffuse_new_particle(sim,top,start,vel);
      }

      /* replace the particles that are not in the correct cell */
      /* it doesn't recalculate the proper number of particles per cell */
      if (retval < 0)
         top = build_new_tree(file,sim,top);

      /* remove any cells with no particles, and combine cells with
       * too few particles, it also recalculates the proper number of
       * particles per cell */
      // In DLA, we don't need to do this!
      // clean_up_all_cells(sim,top);
   }

   // close the stats file
   if (file->write_stats) {
      fclose(file->statsfp);
   }

   // return code 1 - ran out of particles
   fprintf(stdout,"Simulation completed.\n");
   return(1);
}


/*
 *  Run the random-walk method until a hit occurs
 */
int diffuse_new_particle(sim_ptr sim,cell_ptr top,FLOAT* start,FLOAT* vel) {

   int retval,d,trying_particles,still_walking;
   int num_particles_tried = 0;	// how many particles did we try?
   int num_steps_taken = 0;	// how many steps did it take to stick?
   FLOAT loc[DIM],center[DIM],rad,initvel[DIM];
   FLOAT overlap,moveaway;
   particle_ptr newpart;	// pointer to the new particle
   particle_ptr closepart;	// pointer to particle that newpart stuck to

   // overlap is non-dimensional
   overlap = 2.0*(1.0-sim->penetration);

   // this is dimensional
   moveaway = (2.0+sim->penetration)*sim->new_part_rad;

   // use negative velocity for initial particle placement?
   // for (j=0;j<DIM;j++) initvel[j]=-1.0*vel[j];
   for (d=0;d<DIM;d++) initvel[d]=0.0;

   // create the new part right here
   // start[3] comes in as zeroes
   newpart = new_stationary_particle(sim->step,sim->new_part_rad,start);

   trying_particles = TRUE;
   while (trying_particles) {

      num_particles_tried++;

      // find particle start point
      for (d=0;d<DIM;d++) loc[d]=start[d];

      // take one step to start
      rad = find_dist_to_farthest(top,loc,0.0)+2.0*overlap*sim->new_part_rad;
      // rad = find_dist_to_closest(top,loc,LARGE,newpart,&closepart)-2.0*sim->new_part_rad;
      sim->overall_rad = rad;
      // fprintf(stderr,"  maxrad is %g\n",rad);
      // fprintf(stderr,"starting at %g %g\n",loc[0],loc[1]);
      for (d=0;d<DIM;d++) center[d]=loc[d];
      pick_point_on_sphere(loc,center,rad-overlap*sim->new_part_rad,FALSE,initvel);
      // fprintf(stderr,"  first move to %g %g\n",loc[0],loc[1]);


      // loop until particle hits a static particle
      still_walking = TRUE;
      while (still_walking) {

         // check for contact between this part and a static one
         closepart = NULL;
         rad = find_dist_to_closest(top,loc,LARGE,newpart,&closepart);
         num_steps_taken++;
         // fprintf(stderr,"  nearest is %g\n",rad);
         // if it's close enough to stick...
         if (rad < 2.0*sim->new_part_rad) {

            // but if the rootward particle is stubborn
            if (sim->use_stubborn && closepart->flag < sim->stubborn) {
               // then don't stick, but make rootward particle less stubborn
               closepart->flag++;
               // it didn't stick, push it away slightly
               // for (d=0;d<DIM;d++) loc[d] += moveaway*(loc[d]-closepart->x[d])/rad;
               for (d=0;d<DIM;d++) loc[d] = closepart->x[d] + moveaway*(loc[d]-closepart->x[d])/rad;

            // or if the rootward particle isn't stubborn anymore...
            } else {

               // see if stickiness allows particle to stick...
               if (sim->use_stickiness) {
                  if ((double)(rand())/(double)(RAND_MAX) < sim->stickiness) {
                     // it sticks!
                     still_walking = FALSE;
                     // fprintf(stderr,"point %g %g stuck to particle at %g %g\n",
                     //   loc[0],loc[1],closepart->x[0],closepart->x[1]);

                     // and if grip is on, slide it closer
                     if (sim->use_grip) {
                        // particle "grips" and slides toward the contacted particle
                        for (d=0;d<DIM;d++) loc[d] -= sim->grip*(loc[d]-closepart->x[d]);
                     }
                  // not this time, pushback and keep walking
                  } else {
                     // fprintf(stderr,"point %g %g didn't stick to particle at %g %g, ",
                     //   loc[0],loc[1],closepart->x[0],closepart->x[1]);
                     // for (d=0;d<DIM;d++) loc[d] += moveaway*(loc[d]-closepart->x[d])/rad;
                     // it didn't stick, push it away slightly
                     for (d=0;d<DIM;d++)
                        loc[d] = closepart->x[d] + moveaway*(loc[d]-closepart->x[d])/rad;
                     // fprintf(stderr,"pushed back to %g %g\n",loc[0],loc[1]);
                  }

               // we always stick
               } else {
                  still_walking = FALSE;
                  // and if grip is on, slide it closer
                  if (sim->use_grip) {
                     // particle "grips" and slides toward the contacted particle
                     for (d=0;d<DIM;d++) loc[d] -= sim->grip*(loc[d]-closepart->x[d]);
                  }
               }

            }
         }

         // if we strayed too far away, quit anyways
         if (rad > 10.*sim->overall_rad) still_walking = FALSE;

         // break out if we should stop now
         if (!still_walking) break;

         // move the particle to a sphere of that radius
         for (d=0;d<DIM;d++) center[d]=loc[d];
         // fprintf(stderr,"  center at %g %g %g\n",center[0],center[1],center[2]);
         pick_point_on_sphere(loc,center,rad-overlap*sim->new_part_rad,sim->use_bulk_vel,vel);
         // fprintf(stderr,"  point at %g %g\n",loc[0],loc[1]);

      }

      // if we quit because we contacted another particle...
      if (rad < 2.0*sim->new_part_rad) {
         // only now locate the particle
         for (d=0;d<DIM;d++) newpart->x[d] = loc[d];
         // return a negative number if the new part is out of bounds
         retval = add_particle_to_cell(sim,newpart,top);
         // fprintf(stderr,"  planting particle at %g %g, rad is %g, npr is %g\n\n",loc[0],loc[1],rad,sim->new_part_rad);

         trying_particles = FALSE;
      } else {
         // or, because we were too far away
         // fprintf(stderr,"  particle at %g %g %g, rad is %g, is too far\n",loc[0],loc[1],loc[2],rad);
      }
   }

   // set statistics
   sim->total_particles_tried += num_particles_tried;
   sim->total_steps_taken += num_steps_taken;

   return(retval);
}


/*
 * find a random point on a given sphere shell
 *
 * inputs
 *   center[DIM]  starting location
 *
 * outputs
 *   loc[DIM]  new point
 */
void pick_point_on_sphere(FLOAT* loc,FLOAT* center,FLOAT rad,int use_vel,FLOAT* vel) {

   // int rotex = TRUE;
   int d,td;
   FLOAT temp[DIM];
   FLOAT len = 2.;
   FLOAT dt = rad*rad;
   FLOAT a[5],sdev[DIM+1],r1,r2;

   // first, account for bulk velocity
   if (use_vel) {
#ifdef ROTEX
      // if (rotex) {
         // point velocity around counterclockwise
         // dt = vec_length_sq(center)*vec_length(center);
         vel[0] = -1000.*center[1];
         vel[1] = 1000.*center[0];
         vel[0] += -1000.*center[0];
         vel[1] += -1000.*center[1];
         for (d=2;d<DIM;d++) vel[d] = 0.;
         // reduce rad so that curves stay curvy
         dt = 0.25*vec_length(center);
         if (rad > dt) rad = dt;
      // }
#endif
      // fprintf(stderr,"bulk vel is %g %g\n",vel[0],vel[1]);
      // first, find the number of standard deviations in each direction
      for (td=0;td<(DIM+1)/2;td++) {
         gaussian_rand(&r1,&r2);
         sdev[2*td] = r1;
         sdev[2*td+1] = r2;
      }
      // fprintf(stderr,"std devs are %g %g\n",sdev[0],sdev[1]);
      // then, set up the quartic equation
      a[4] = vec_length_sq(vel);
      a[3] = 2.0*vec_dot(vel,sdev);
      a[2] = vec_length_sq(sdev);
      a[1] = 0.0;
      a[0] = -1.0*rad*rad;
      // solve it (for sqrt(dt))
      dt = solve_quartic(a);
      // and account for the random motion
      for (d=0;d<DIM;d++) loc[d] = center[d]+dt*sdev[d];
      // determine the real dt
      dt *= dt;
      // move the particle due to convection
      for (d=0;d<DIM;d++) loc[d] += dt*vel[d];
      // now we're done!
      return;
#ifdef OLD
      // determine speed of convection
      len = vec_length(vel);
      // solve for dt such that the particle never leaves a sphere of radius rad
      // WARNING: this is still not quite accurate; how would one determine how
      // long a *specific* particle took to diffuse a specific distance?
      rad = (sqrt(1.0+4.0*rad*len)-1.0)/(2.0*len);
      dt = pow(rad,2);
      // move the particle due to convection
      for (d=0;d<DIM;d++) center[d] += dt*vel[d];
#endif
   }

   // then move it due to random motion
#if DIM==2
#ifdef USE_SINE_SPHERE
   // second method uses no iteration at all, because sin/cos is accelerated
   temp[0] = 6.2831853071795864 * rand()/(RAND_MAX+1.0);
   loc[0] = cos(temp[0]);
   loc[1] = sin(temp[0]);
   for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d];
#else
   // new method needs no sqrt
   len = 2.;
   while (len > 1.) {
      // for (d=0;d<DIM;d++) loc[d] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      for (d=0;d<DIM;d++) temp[d] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      // len = vec_length(loc);
      len = temp[0]*temp[0]+temp[1]*temp[1];
   }
   // for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d]/len;
   loc[0] = (temp[0]*temp[0]-temp[1]*temp[1])/len;
   loc[1] = 2.*temp[0]*temp[1]/len;
   // printf("pt on sphere %g %g\n",loc[0],loc[1]);
   for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d];
   // printf("scaled pt %g %g\n",loc[0],loc[1]);
#endif
#elif DIM==3
   // new method needs no sqrt
   len = 2.;
   while (len > 1.) {
      // for (d=0;d<DIM;d++) loc[d] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      for (d=0;d<DIM;d++) temp[d] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      // len = vec_length(loc);
      len = temp[0]*temp[0]+temp[1]*temp[1];
   }
   // printf("pt on circle %g %g\n",temp[0],temp[1]);
   // for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d]/len;
   // loc[0] = (temp[0]*temp[0]-temp[1]*temp[1])/len;
   // loc[1] = 2.*temp[0]*temp[1]/len;
   loc[2] = 2.*len -1.;
   len = 2.*sqrt(1.-len);
   loc[0] = temp[0]*len;
   loc[1] = temp[1]*len;
   // printf("pt on sphere %g %g %g\n",loc[0],loc[1],loc[2]);
   for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d];
   // printf("new pt %g %g %g\n",loc[0],loc[1],loc[2]);
#else
   // else use the standard method
   len = 2.;
   while (len > 1.) {
      for (d=0;d<DIM;d++) loc[d] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      len = vec_length_sq(loc);
   }
   len = sqrt(len);
   for (d=0;d<DIM;d++) loc[d] = center[d]+rad*loc[d]/len;
#endif

   return;
}


/*
 *  Find the distance to the closest particle
 *  based on is_this_space_open2
 *
 *  cell is cell to search within
 *  loc is point to search against
 *  nearest is current value for nearest
 */
// FLOAT find_dist_to_closest(cell_ptr cell,FLOAT loc[DIM],FLOAT neardist,
FLOAT find_dist_to_closest(cell_ptr cell,FLOAT* loc,FLOAT neardist,
                           particle_ptr newpart,particle_ptr *closest){

   int i,d;
   int s[DIM],sstart;
   // int need_to_do_sqrt = FALSE;
   FLOAT temp[DIM],dist,nsq;
   particle_ptr curr;

   nsq = neardist*neardist;

   // are all corners of this cell 
   for (d=0;d<DIM;d++) {
      if (loc[d] < cell->min[d]-neardist) return(neardist);
      if (loc[d] > cell->max[d]+neardist) return(neardist);
   }

   if (cell->has_subcells) {
      // fprintf(stdout,"    cell has %d subcells\n",NCHILD);
      // must check all of the subcells, but check closest first
      for (d=0;d<DIM;d++) if (loc[d] < cell->mid[d]) s[d]=0; else s[d]=1;
      sstart = 0;
      for (d=0;d<DIM;d++) sstart += s[d]*pow(2,d);
      for (i=sstart;i<sstart+NCHILD;i++) {
      // the slow way does in 27 sec what the fast way does in 19
      // for (i=0;i<NCHILD;i++) {
         // dist = find_dist_to_closest(cell->s[i],
         dist = find_dist_to_closest(cell->s[i%NCHILD],
                   loc,neardist,newpart,closest);
         if (dist < neardist) neardist = dist;
         // fprintf(stderr,"    closest is %g\n",neardist);
      }
   } else {
      // fprintf(stdout,"    cell has %d particles\n",cell->num);
      // check vs. all of the particles in this cell
      // fprintf(stderr,"    checking vs %d particles in box at level %d\n",cell->num,cell->level);
      curr = cell->first;
      while (curr) {

         // it DOES NOT PAY to test this sum at every addition!
         for (d=0;d<DIM;d++) temp[d] = loc[d]-curr->x[d];
         dist = vec_length_sq(temp);

         // fprintf(stderr,"    dist to this particle is %g\n",dist);
         if (dist < nsq) {
            nsq = dist;
            // it DOES NOT PAY to postpone the sqrt until later!
            neardist = sqrt(dist);
            // need_to_do_sqrt = TRUE;
            (*closest) = curr;
            newpart->root = curr;
         }
         curr = curr->next;
      }
   }

   // if (need_to_do_sqrt) neardist = sqrt(dist);

   return(neardist);
}


/*
 *  Find the distance to the farthest particle
 *  based on is_this_space_open2
 */
// FLOAT find_dist_to_farthest(cell_ptr cell,FLOAT loc[DIM],FLOAT farthest){
FLOAT find_dist_to_farthest(cell_ptr cell,FLOAT* loc,FLOAT farthest){

   int i,d;
   FLOAT temp[DIM],dsq,fsq,dmin,dmax,dist;
   particle_ptr curr;

   // if (cell->level == 0) farthest = 0.;
   // farthest = 0.;
   fsq = farthest*farthest;

   // must see if whole cell is inside of sphere radius "farthest"
   // so, find the distance to the farthest corner on the cube
   dsq = 0.;
   for (d=0;d<DIM;d++) {
      dmin = fabs(loc[d]-cell->min[d]);
      dmax = fabs(loc[d]-cell->max[d]);
      if (dmin > dmax) dsq += dmin*dmin;
      else dsq += dmax*dmax;
   }
   if (dsq < fsq) return(farthest);

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<NCHILD;i++) {
         dist = find_dist_to_farthest(cell->s[i],loc,farthest);
         if (dist > farthest) farthest = dist;
         // fprintf(stderr,"    farthest is %g\n",farthest);
      }
   } else {
      // check vs. all of the particles in this cell
      // fprintf(stderr,"    checking vs %d particles in box at level %d\n",cell->num,cell->level);
      curr = cell->first;
      while (curr) {
         for (d=0;d<DIM;d++) temp[d] = loc[d]-curr->x[d];
         dist = vec_length_sq(temp);
         // fprintf(stderr,"    dist to this particle is %g\n",dist);
         if (dist > fsq) {
            fsq = dist;
            farthest = sqrt(dist);
         }
         curr = curr->next;
      }
   }

   return(farthest);
}


/*
 * Starting with each particle, march toward the center
 * and add 1.0=mass to every node along the way
 */
void sum_tipward_mass(cell_ptr cell) {

   int i;
   particle_ptr curr,march;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<NCHILD;i++)
         sum_tipward_mass(cell->s[i]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->mass += 1.0;
         march = curr->root;
         while (march) {
            march->mass += 1.0;
            march = march->root;
         }
         curr = curr->next;
      }
   }

   return;
}


/*
 * zero all masses
 */
void zero_all_masses(cell_ptr cell) {

   int i;
   particle_ptr curr;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<NCHILD;i++)
         zero_all_masses(cell->s[i]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->mass = 0.0;
         curr = curr->next;
      }
   }

   return;
}


/*
 * zero all flags - this is not normally needed
 */
void zero_flags(cell_ptr cell) {

   int i;
   particle_ptr curr;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<NCHILD;i++)
         zero_flags(cell->s[i]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->flag = 0;
         curr = curr->next;
      }
   }

   return;
}


/*
 * increment all flags - this is not normally needed
 */
void increment_flags(cell_ptr cell) {

   int i;
   particle_ptr curr;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<NCHILD;i++)
         increment_flags(cell->s[i]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->flag++;
         curr = curr->next;
      }
   }

   return;
}


/*
 * find the 2-norm of a d-length vector
 */
FLOAT vec_dot(FLOAT* x,FLOAT* y) {

   int i;
   FLOAT sum = 0.0;

   for (i=0;i<DIM;i++) sum += x[i]*y[i];

   return(sum);
}


/*
 * find the 2-norm of a d-length vector
 */
FLOAT vec_length(FLOAT* x) {

   int i;
   FLOAT sum = 0.0;

   for (i=0;i<DIM;i++) sum += x[i]*x[i];
   sum = sqrt(sum);

   return(sum);
}


/*
 * find the 2-norm squared of a d-length vector
 */
FLOAT vec_length_sq(FLOAT* x) {

   int i;
   FLOAT sum = 0.0;

   for (i=0;i<DIM;i++) sum += x[i]*x[i];

   return(sum);
}


/*
 * return two random numbers from the normal distribution, mean=0, sigma=1
 *
 * uses the Box-Mueller transformation
 */
void gaussian_rand(FLOAT* r1,FLOAT* r2) {

   double t1,t2;

   t1 = (double)(rand())/(double)(RAND_MAX);
   t2 = (double)(rand())/(double)(RAND_MAX);

   *r1 = (FLOAT)(sqrt(-2.*log(t1))*cos(2.*M_PI*t2));
   *r2 = (FLOAT)(sqrt(-2.*log(t1))*sin(2.*M_PI*t2));

   return;
}


/*
 * solve the quartic equation numerically for the real positive root;
 * equation is a[4] x^4 + a[3] x^3 + a[2] x^2 + a[1] x + a[0] = 0
 */
FLOAT solve_quartic(FLOAT* a) {

   int i,it,maxit = 100;
   FLOAT xmult;
   FLOAT bottomx,bottomval;
   FLOAT topx,topval;
   FLOAT newx,newval;

   // evaluate the function at the bottom bound (always negative in our case)
   bottomx = 0.0;
   bottomval = a[0];
   // fprintf(stderr,"initial bottom bound f(%g)=%g\n",bottomx,bottomval);

   // evaluate the function at the top bound
   topx = fabs((a[3]+a[0])/2.0*a[4]);
   topval = -1.0;
   while (topval < 0.0) {
      topx *= 2.0;
      topval=a[0];
      xmult=1.;
      for (i=1;i<5;i++) {
         xmult *= topx;
         topval += a[i]*xmult;
      }
      // fprintf(stderr,"initial top bound f(%g)=%g\n",topx,topval);
      // we hope the top bound is positive
   }

   newval = 1.0;
   it = 0;
   while (fabs(newval) > 1.0e-6 && it < maxit) {
      // take a new guess for x
      newx = (topx+bottomx)/2.0;
      // evaluate the expression
      newval=a[0];
      xmult=1.;
      for (i=1;i<5;i++) {
         xmult *= newx;
         newval += a[i]*xmult;
      }
      // fprintf(stderr,"   interim value f(%g)=%g\n",newx,newval);
      if (newval < 0.0) {
         // replace the bottom value if negative
         bottomx = newx;
      } else {
         // replace the top value if positive
         topx = newx;
      }
      it++;
   }

   return newx;
}

