/*************************************************************
 *
 *  setup.c - subroutines for setting up simulations
 *
 *  Copyright (C) 2000-2005  Mark J. Stock, mstock@umich.edu
 *
 *  This file is part of dla-nd.
 *
 *  dla-nd is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  dla-nd is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dla-nd; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *********************************************************** */


#include "structs.h"
#include <ctype.h>

int initialize_system(fileprop_ptr,sim_ptr,cell_ptr);
int add_particles_from_file(sim_ptr,cell_ptr,int,char);
particle_ptr new_particle(int,FLOAT,FLOAT,FLOAT*,FLOAT*);
int read_input_file(fileprop_ptr,sim_ptr,cell_ptr);
int parse_args(int,char**,fileprop_ptr);
int Usage(char[80],int);
FLOAT*** allocate_3d_array_F(int,int,int);
FLOAT** allocate_2d_array_F(int,int);
png_byte** allocate_2d_array_pb(int,int,int);


/*
 *  Initialize the system - make static and free element structures, etc.
 */
int initialize_system(fileprop_ptr file,sim_ptr sim,cell_ptr top) {

   int i,j,k,d,ret_val,num_parts;
   FLOAT loc[DIM],vel[DIM],xpos,ypos,zpos,xvel,yvel,zvel,distsq,mpp;
   FLOAT r,rcubed,orbvel,M;
   particle_ptr newpart;

   // -------------------------------------------------------------------------

   /* set default values for variables */
   // strcpy(file->out_fn_root,"output_");
   strcpy(file->out_fn_root,"");
   file->write_pgm = FALSE;
   file->write_png = TRUE;
   file->write_dot = FALSE;
   file->write_rad = FALSE;
   file->write_obj = FALSE;
   file->write_part = FALSE;
   file->image_depth = 8;
   file->out_img_size = 256;

   file->write_stats = TRUE;
   strcpy(file->stats_file,"stats.dat");

   file->write_temp = TRUE;
   strcpy(file->temp_file,".temp.p3d");

   sim->seed = 1;
   sim->step = 0;
   sim->output_step = 100;
   sim->max_levels = 50;
   sim->max_parts_in_cell = 40;

   sim->next_output_step = 0;
   sim->next_output_index = 1;

   sim->penetration = 0.1;

   sim->use_stickiness = FALSE;
   sim->stickiness = 1.0;

   sim->use_grip = FALSE;
   sim->grip = 0.0;

   sim->use_stubborn = FALSE;
   sim->stubborn = 0;

   sim->use_bulk_vel = FALSE;
   for (d=0;d<DIM;d++) {
      sim->bulk_vel[d] = 0.0;
   }

   for (d=0;d<DIM;d++) {
      sim->bdry[d][0] = WALL;
      sim->bdry[d][1] = WALL;
   }

   sim->particle_cnt = 1000;
   sim->new_part_rad = 0.005;
   sim->new_part_mass = 1.0;

   sim->num_blocks = 0;
   sim->num_spheres = 0;
   sim->num_indiv_parts = 0;
   sim->num_read_part = 0;
   sim->num_read_stat = 0;
   sim->use_density_field = FALSE;
   sim->construct_dens_surface = FALSE;


   /* assign default values for the top-level cell */
   top->level = 0;
   top->has_subcells = FALSE;
   top->first = NULL;
   for (i=0;i<NCHILD;i++)
      top->s[i] = NULL;
   for (d=0;d<DIM;d++) {
      top->min[d] = 0.0;
      top->max[d] = 1.0;
   }
   top->num = 0;
   top->mass = 0.0;

   // -------------------------------------------------------------------------

   /* now, if the file listed on the command line exists, read it */
   /* ret_val is 0 if file is OK and read, 1 if not */
   ret_val = read_input_file(file,sim,top);

   if (ret_val == 1) {
      fprintf(stdout,"ERROR upon reading input file, please check syntax\n");
      exit(1);
   }

   // -------------------------------------------------------------------------

   // "seed" the random number generator
   srand(sim->seed);

   // cell centers are for tree-traversing
   for (d=0;d<DIM;d++) top->mid[d] = (top->min[d]+top->max[d])/2.0;

   // I think we only need these for the self-gravitation thing
   for (d=0;d<DIM;d++) top->cm[d] = (top->min[d]+top->max[d])/2.0;

   // -------------------------------------------------------------------------
   // now, place the seed particle(s)

   for (i=0; i<sim->num_read_stat; i++)
      add_particles_from_file(sim,top,i,TRUE);

   // then, place specific particles
   for (i=0; i<sim->num_indiv_parts; i++) {
      for (d=0;d<DIM;d++) loc[d] = sim->indiv_part[i][d+1];
      for (d=0;d<DIM;d++) vel[d] = sim->indiv_part[i][d+DIM+1];
      newpart = new_particle(i,sim->indiv_part[i][0],sim->new_part_rad,loc,vel);
      add_particle_to_cell(sim,newpart,top);
   }

   // -------------------------------------------------------------------------

   if (sim->use_density_field) {
      // set the cell sizes
      for (i=0; i<2; i++) sim->ff2->n[i] = file->out_img_size;
      for (i=0; i<2; i++) sim->ff2->d[i] = (top->max[i]-top->min[i])/sim->ff2->n[i];
      // fprintf(stdout,"ff2 has %d %d cells, each %g by %g\n",sim->ff2->n[0],sim->ff2->n[1],sim->ff2->d[0],sim->ff2->d[1]);

      // allocate memory for the array
      // ff->rho = allocate_3d_array_F(ff->n[0],ff->n[1],ff->n[2]);
      // sim->ff2->rho = allocate_2d_array_F(sim->ff2->n[0],sim->ff2->n[2]);
      sim->ff2->rho = allocate_2d_array_F(sim->ff2->n[0],sim->ff2->n[1]);

      // zero the array
      for (i=0;i<sim->ff2->n[0];i++)
         for (j=0;j<sim->ff2->n[1];j++)
            sim->ff2->rho[i][j] = 0.0;
            // for (k=0;k<sim->ff2->n[2];k++)
               // ff->rho[i][j][k] = 0.0;
   }

   // fprintf(stdout,"\nStarting with %d particles",top->num);

   // and the output file (just dots)
   for (i=0; i<2; i++) file->out->d[i] = (top->max[i]-top->min[i])/file->out->n[i];
   if (file->write_dot || file->write_png)
      file->out->rho = allocate_2d_array_F(file->out_img_size,file->out_img_size);

   // malloc the space for png writing, if necessary
   if (file->write_png)
      file->image = allocate_2d_array_pb(file->out_img_size,file->out_img_size,file->image_depth);

   /* return 0, everything OK */
   return(0);
}


/*
 *  Read particle locations from a file, include them in the simulation
 *  Do this for both static and non-static particles
 */
int add_particles_from_file(sim_ptr sim,cell_ptr top,int i,char is_stat) {

   int d;
   int cnt = 0;
   FLOAT loc[DIM],rad,vel[DIM];
   char token[7][16];
   char twochar[2];
   char sbuf[512];
   particle_ptr newpart;
   FILE *infile;

   /* open file for reading */
   if (is_stat) {
      infile = fopen(sim->read_stat[i],"r");
      if (infile==NULL) {
         fprintf(stderr,"Could not open particle file %s\n",sim->read_stat[i]);
         fflush(stderr);
         return(1);
      }
      fprintf(stdout,"Opening file %s\n",sim->read_stat[i]);
      fflush(stdout);
   } else {
      infile = fopen(sim->read_part[i],"r");
      if (infile==NULL) {
         fprintf(stderr,"Could not open particle file %s\n",sim->read_part[i]);
         fflush(stderr);
         return(1);
      }
      fprintf(stdout,"Opening file %s\n",sim->read_part[i]);
      fflush(stdout);
   }

   /* read a line from the input file */
   while (fscanf(infile,"%[^\n]",sbuf) != EOF) {

      /* fprintf(stdout,"%s\n",sbuf); */

      /* grab the line */
      sscanf(sbuf,"%s %s %s %s %s %s %s",token[0],token[1],token[2],token[3],token[4],token[5],token[6]);
      /* fprintf(stdout,"   first token is %s\n",token[0]); */

      if (token[0][0] == '#') {
         /* read a comment line, or some other descriptor */
         fscanf(infile,"%[\n]",twochar);    /* read up to newline */
         /* fprintf(stdout,"%s\n",sbuf);        /* write comment */

      } else {
         /* read the particle parameters */
         /* fprintf(stdout,"   second token is %s\n",token[1]); */

         for (d=0;d<DIM;d++) loc[d] = atof(token[d]);
         rad = atof(token[DIM]);

         for (d=0;d<DIM;d++) {
            loc[d] += sim->stat_xform[i][d];
            vel[d] = 0.0;
         }
         newpart = new_particle(cnt,sim->new_part_mass,rad,loc,vel);

         add_particle_to_cell(sim,newpart,top);

         cnt++;
         if (cnt/1000 != (cnt+1)/1000) {
            fprintf(stdout,".");
            fflush(stdout);
         }

         fscanf(infile,"%[\n]",twochar);    /* read newline */
      }
   }

   fprintf(stdout,"\n");
   fprintf(stdout,"Placed %d particles\n",cnt);
   fflush(stdout);

   fclose(infile);

   /* ret_val is 0 is file is OK and read, 1 if not */
   return(0);
}


/*
 *  Create a new particle
 */
particle_ptr new_particle(int i,FLOAT m,FLOAT r,FLOAT* x,FLOAT* u){

   int d;
   particle_ptr newpart;

   newpart = (PARTICLE*)malloc(sizeof(PARTICLE));
   newpart->index = i;
   newpart->mass = m;
   newpart->rad = r;
   for (d=0;d<DIM;d++) newpart->x[d] = x[d];
   // for (d=0;d<DIM;d++) newpart->u[d] = u[d];
   newpart->flag = 0;
   newpart->stationary = TRUE;
   newpart->next = NULL;
   newpart->root = NULL;

   // fprintf(stderr,"Added particle %d with mass %g and radius %g\n",newpart->index,newpart->mass,newpart->rad);

   return(newpart);
}


/*
 *  Create a new stationary particle
 */
particle_ptr new_stationary_particle(int i,FLOAT r,FLOAT* x){

   int d;
   particle_ptr newpart;

   newpart = (PARTICLE*)malloc(sizeof(PARTICLE));
   newpart->index = i;
   newpart->mass = 0.0;
   newpart->rad = r;
   for (d=0;d<DIM;d++) newpart->x[d] = x[d];
   newpart->flag = 0;
   newpart->stationary = TRUE;
   newpart->next = NULL;
   newpart->root = NULL;

   return(newpart);
}


/*
 * Read the input file and set the listed parameters
 */
int read_input_file(fileprop_ptr file,sim_ptr sim,cell_ptr top) {

   int d;
   char token[13][32];
   char twochar[2];
   char sbuf[512];
   FILE *infile;

   if (file->use_input_file) {

      // open file for reading
      infile = fopen(file->input_fn,"r");
      if (infile==NULL) {
         fprintf(stderr,"Could not open input file %s\n",file->input_fn);
         fflush(stderr);
         return(1);
      }
      fprintf(stdout,"Opening file %s\n",file->input_fn);
      fflush(stdout);

   } else {

      // open stdin for reading
      infile = stdin;
      fprintf(stdout,"Opening stdin for reading...\n");
      fflush(stdout);
   }


   /* read a line from the input file */
   while (fscanf(infile,"%[^\n]",sbuf) != EOF) {

      // fprintf(stdout,"%s\n",sbuf);

      /* grab the first word */
      sscanf(sbuf,"%s %s %s %s %s %s %s %s %s %s %s %s %s",token[0],token[1],token[2],token[3],token[4],token[5],token[6],token[7],token[8],token[9],token[10],token[11],token[12]);
      /* fprintf(stdout,"   first token is %s\n",token[0]); */

      if (token[0][0] == '#') {
         /* read a comment line, or some other descriptor */
         fscanf(infile,"%[\n]",twochar);    /* read up to newline */
         /* fprintf(stdout,"%s\n",sbuf);        /* write comment */

      } else {
         /* first word is a parameter name */
         /* read the parameter value */
         /* fprintf(stdout,"   second token is %s\n",token[1]); */

         if (strncmp(token[0],"outfile_root",12) == 0) {
            strcpy(file->out_fn_root,token[1]);
         } else if (strncmp(token[0],"particles",9) == 0) {
            sim->particle_cnt = atoi(token[1]);
         } else if (strncmp(token[0],"write_pgm",9) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_pgm = TRUE;
            else
               file->write_pgm = FALSE;
         } else if (strncmp(token[0],"write_png",9) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_png = TRUE;
            else
               file->write_png = FALSE;
         } else if (strncmp(token[0],"write_dot",9) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_dot = TRUE;
            else
               file->write_dot = FALSE;
         } else if (strncmp(token[0],"write_obj",9) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_obj = TRUE;
            else
               file->write_obj = FALSE;
         } else if (strncmp(token[0],"write_rad",9) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_rad = TRUE;
            else
               file->write_rad = FALSE;
         } else if (strncmp(token[0],"write_part",10) == 0) {
            if (strncmp(token[1],"yes",1) == 0)
               file->write_part = TRUE;
            else
               file->write_part = FALSE;
         } else if (strncmp(token[0],"write_stats",11) == 0) {
            file->write_stats = TRUE;
            if (isalnum(token[1][0])) {
               strcpy(file->stats_file,token[1]);
            }
         } else if (strncmp(token[0],"write_temp",10) == 0) {
            file->write_temp = TRUE;
            if (isalnum(token[1][0])) {
               strcpy(file->temp_file,token[1]);
            }
         } else if (strncmp(token[0],"output_step",9) == 0) {
            sim->output_step = atoi(token[1]);
         } else if (strncmp(token[0],"max_levels",10) == 0) {
            sim->max_levels = atoi(token[1]);
         } else if (strncmp(token[0],"max_ppc",7) == 0) {
            sim->max_parts_in_cell = atoi(token[1]);
         } else if (strncmp(token[0],"seed",4) == 0) {
            sim->seed = atoi(token[1]);
         } else if (strncmp(token[0],"penetration",5) == 0) {
            sim->penetration = atof(token[1]);
         } else if (strncmp(token[0],"stickiness",5) == 0) {
            sim->use_stickiness = TRUE;
            sim->stickiness = atof(token[1]);
         } else if (strncmp(token[0],"comp_domain",11) == 0) {
            /* set initial size of first cell */
            for (d=0;d<DIM;d++) {
               top->min[d] = atof(token[2*d+1]);
               top->max[d] = atof(token[2*d+2]);
            }
         } else if (strncmp(token[0],"boundary",8) == 0) {
            if (strncmp(token[1],"wall",4) == 0) {
               for (d=0;d<DIM;d++) {
                  sim->bdry[d][0] = WALL;
                  sim->bdry[d][1] = WALL;
               }
            } else if (strncmp(token[1],"open",4) == 0) {
               for (d=0;d<DIM;d++) {
                  sim->bdry[d][0] = OPEN;
                  sim->bdry[d][1] = OPEN;
               }
            }
         } else if (strncmp(token[0],"image_size",10) == 0) {
            file->out_img_size = atoi(token[1]);
            for (d=0;d<DIM;d++) file->out->n[d] = atoi(token[1]);
            // fprintf(stdout,"img size is %d\n",file->out_img_size);
         } else if (strncmp(token[0],"image_depth",11) == 0) {
            file->image_depth = atoi(token[1]);
         } else if (strncmp(token[0],"rad",3) == 0) {
            sim->new_part_rad = atof(token[1]);
         } else if (strncmp(token[0],"mass",4) == 0) {
            sim->new_part_mass = atof(token[1]);
         } else if (strncmp(token[0],"add_part",8) == 0) {
            sim->indiv_part[sim->num_indiv_parts][0] = atof(token[1]);
            for (d=1;d<2*DIM+1;d++) {
               sim->indiv_part[sim->num_indiv_parts][d] = atof(token[d+1]);
            }
            sim->num_indiv_parts++;
         } else if (strncmp(token[0],"read_part",9) == 0) {
            strcpy(sim->read_part[sim->num_read_part],token[1]);
            for (d=0;d<2*DIM;d++)
               sim->part_xform[sim->num_read_part][d] = atof(token[d+2]);
            sim->num_read_part++;
         } else if (strncmp(token[0],"read_stat",9) == 0) {
            strcpy(sim->read_stat[sim->num_read_stat],token[1]);
            for (d=0;d<DIM;d++)
               sim->stat_xform[sim->num_read_stat][d] = atof(token[d+2]);
            sim->num_read_stat++;
         } else if (strncmp(token[0],"write_dens",10) == 0) {
            // sim->ff2->n[0] = atoi(token[1]);
            // sim->ff2->n[1] = atoi(token[1]);
            // sim->ff2->n[2] = atoi(token[1]);
            // sim->ff3->n[0] = atoi(token[1]);
            // sim->ff3->n[1] = atoi(token[1]);
            // sim->ff3->n[2] = atoi(token[1]);
            // sim->use_density_field = TRUE;
            if (strncmp(token[1],"yes",1) == 0)
               sim->use_density_field = TRUE;
            else
               sim->use_density_field = FALSE;
         } else if (strncmp(token[0],"grip",4) == 0) {
            sim->use_grip = TRUE;
            sim->grip = atof(token[1]);
         } else if (strncmp(token[0],"stub",4) == 0) {
            sim->use_stubborn = TRUE;
            sim->stubborn = atoi(token[1]);
         } else if (strncmp(token[0],"bulk_vel",8) == 0) {
            sim->use_bulk_vel = TRUE;
            for (d=0;d<DIM;d++) sim->bulk_vel[d] = atof(token[d+1]);
         }

         fscanf(infile,"%[\n]",twochar);    /* read newline */
      }
   }

   fclose(infile);

   /* ret_val is 0 is file is OK and read, 1 if not */
   return(0);
}


/*
 * Parse the command-line arguments
 */
int parse_args(int argc,char **argv,fileprop_ptr file) {

   int i,some_logical;
   FLOAT some_float;

   // copy executable name
   (void) strcpy(file->exectuable_fn,argv[0]);

   if (argc < 2) {

      // if no command-line arg, then input must be stdin
      file->use_input_file = FALSE;

   } else {

      // if command-line arg is "help", write usage
      if (strncmp(argv[1], "--help", 3) == 0)
         (void) Usage(file->exectuable_fn,0);
      if (strncmp(argv[1], "-help", 2) == 0)
         (void) Usage(file->exectuable_fn,0);
      if (strncmp(argv[1], "help", 4) == 0)
         (void) Usage(file->exectuable_fn,0);

      // if not, though, save it as the input file name
      (void) strcpy(file->input_fn,argv[1]);
      file->use_input_file = TRUE;

   }

   return(0);
}


/*
 * This function writes basic usage information to stderr,
 * and then quits. Too bad.
 */
int Usage(char progname[80],int status) {

   /* Usage for grav3d */
   static char **cpp, *help_message[] =
   {
       "  The input file should be a raw ASCII file containing 2 columns, the first",
       "  with a parameter name (dt,particles,G,write_rad) and the second with either",
       "  an integer or floating-point value or with a \'yes\' or \'no\'.",
       " ",
       "  There is running output to stdout, but PGM and Radiance data are written",
       "  to disk.",
       " ",
       NULL
   };

   fprintf(stderr,"\n  Usage:  %s infile\n\n", progname);
   for (cpp = help_message; *cpp; cpp++)
      fprintf(stderr, "%s\n", *cpp);
      fflush(stderr);
   exit(status);
   return(0);
}


/*
 * allocate memory for a three-dimensional array of FLOATs
 */
FLOAT*** allocate_3d_array_F(int nx, int ny, int nz) {

   int i,j;
   FLOAT ***array = (FLOAT ***)malloc(nx * sizeof(FLOAT **));

   array[0] = (FLOAT **)malloc(nx * ny * sizeof(FLOAT *));
   array[0][0] = (FLOAT *)malloc(nx * ny * nz * sizeof(FLOAT));

   for (i=1; i<nx; i++)
      array[i] = array[0] + i * ny;

   for (i=0; i<nx; i++) {
      if (i!=0)
         array[i][0] = array[0][0] + i * ny * nz;
      for (j=1; j<ny; j++)
         array[i][j] = array[i][0] + j * nz;
   }

   return(array);
}


/*
 * allocate memory for a two-dimensional array of FLOATs
 */
FLOAT** allocate_2d_array_F(int nx, int ny) {

   int i,j;
   FLOAT **array = (FLOAT **)malloc(nx * sizeof(FLOAT *));

   array[0] = (FLOAT *)malloc(nx * ny * sizeof(FLOAT));

   for (i=1; i<nx; i++)
      array[i] = array[0] + i * ny;

   return(array);
}


/*
 * allocate memory for a two-dimensional array of png_byte
 */
png_byte** allocate_2d_array_pb(int nx, int ny, int depth) {

   int i,j,bytesperpixel;
   png_byte **array;

   if (depth <= 8) bytesperpixel = 1;
   else bytesperpixel = 2;
   array = (png_byte **)malloc(ny * sizeof(png_byte *));
   array[0] = (png_byte *)malloc(bytesperpixel * nx * ny * sizeof(png_byte));

   for (i=1; i<ny; i++)
      array[i] = array[0] + i * bytesperpixel * nx;

   return(array);
}
